from unittest import TestCase
import numpy as np
from osirisl1services.readlevel1 import Orbit, open_level1_attitude, open_level1_spectrograph, open_level1_ir
import osl1 as l1
from osirisl1services.spatialrotation import ECI, ECEF, Geodetic
import sasktran as sk
from osirisl1services.services import (get_spectrograph_pointing_for_attitude_dataset, get_spectrograph_pointing,
                                       get_pointing_for_spectrograph_dataset,
                                       get_position_for_spectrograph_dataset,
                                       get_tangent_point_data_for_spectrograph_dataset,
                                       get_pointing_for_ir_dataset, get_position_for_ir_dataset,
                                       get_tangent_point_data_for_ir_dataset,
                                       get_scannos_in_orbit_below_sza,
                                       SPECTROGRAPH_ROTATION_TB_ZYX, IR_ROTATION_TB_ZYX)
from osirisl1services.psf_fit import PointSpreadFunctionFit


class TestLevel1Services(TestCase):
    def setUp(self):
        self.scannos = [6432013,  # catches #25
                        27110030]  # 27110030 catches #25

    def test_spectrograph_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/osoptics.ini

            Vector(x=0.99999495087295287, y=-0.0027594116765072515, z=0.0015760316621056657)
        """
        osiris_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.OS_X_BORE)[1]
        q = SPECTROGRAPH_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), osiris_look_cf)

    def test_level1_spectrograph_pointing(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            attitude['osiris_look_eci'] = get_spectrograph_pointing_for_attitude_dataset(attitude['Qachieved'])
            for mjd, att in attitude.groupby('mjd'):
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                np.testing.assert_allclose(att['osiris_look_eci'], osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_spectrograph_pointing_with_interpolation(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            for mjd in np.linspace(attitude.mjd.min(), attitude.mjd.max(), 10, endpoint=True):
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                look = get_spectrograph_pointing(mjd, attitude['Qachieved'])
                np.testing.assert_allclose(look, osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_pointing_for_spectrograph_dataset(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            spectrograph = open_level1_spectrograph(scanno=scanno)
            spectrograph['look_eci'] = get_pointing_for_spectrograph_dataset(spectrograph,
                                                                             attitude['Qachieved'])
            scan_info_swig, _, _, _, _ = l1.LoadScanOS(scanno)
            for info in scan_info_swig:
                mjd = info.mjd
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                np.testing.assert_allclose(spectrograph['look_eci'].sel(mjd=mjd), osiris_look_eci, atol=1e-6,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_spectrograph_SSC_tangent_point(self):
        """ Test comparing the tangent point derived by the attitude file to the tangent point returned
            by the c++ Level1 Services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            attitude['osiris_look_eci'] = get_spectrograph_pointing_for_attitude_dataset(attitude['Qachieved'])
            for mjd, att in attitude.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                position_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(position_eci))[1])
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                osiris_look_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(osiris_look_eci))[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci), list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])

                tangent_point_sasktran = sk.LineOfSight(mjd=mjd,
                                                        observer=position_ecef,
                                                        look_vector=osiris_look_ecef).tangent_location()

                self.assertAlmostEqual(tangentpoint.latitude, tangent_point_sasktran.latitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.longitude, tangent_point_sasktran.longitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.height * 1000.0, tangent_point_sasktran.altitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')

    def test_get_tangent_point_data_for_spectrograph_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the spectrograph.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            spectrograph = open_level1_spectrograph(scanno=scanno)
            # Load the look vector and transform it to ECI
            spectrograph['look_eci'] = get_pointing_for_spectrograph_dataset(spectrograph, attitude['Qachieved'])
            spectrograph['position_eci'] = get_position_for_spectrograph_dataset(spectrograph, attitude)
            spectrograph.merge(get_tangent_point_data_for_spectrograph_dataset(spectrograph, attitude), inplace=True)
            for mjd, spect in spectrograph.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                # Assert the ECI positions are the to 5 meters
                np.testing.assert_allclose(spect['position_eci'], position_eci, atol=5)
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])

                self.assertAlmostEqual(float(spect['latitude']), tangentpoint.latitude, delta=0.01,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(float(spect['longitude']), tangentpoint.longitude, delta=0.01,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(float(spect['altitude']), tangentpoint.height * 1000.0, delta=5,
                                       msg=f'ScanNo: {scanno}')

    def test_get_tangent_point_data_for_spectrograph_dataset_prunes_nans(self):
        """ Test that tangent points of NaNs are pruned from the returned tangent points. """
        scanno = 51236096
        attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
        spectrograph = open_level1_spectrograph(scanno=scanno)
        tangent_point = get_tangent_point_data_for_spectrograph_dataset(spectrograph=spectrograph, attitude=attitude)
        self.assertLess(tangent_point.sizes['mjd'], spectrograph.sizes['mjd'],
                        msg=f'NaN tangent points were not pruned from ScanNo: {scanno}.')


    def test_ir_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/iroptics.ini

            Vector(x=0.999999420741261, y=0.0, z=0.0010763443418775663)
        """
        ir1_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR1_X_BORE)[1]
        ir2_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR2_X_BORE)[1]
        ir3_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR3_X_BORE)[1]
        q = IR_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), ir1_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir2_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir3_look_cf)

    def pointing_for_ir_channel_dataset(self, channel: int):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            ir = open_level1_ir(scanno=scanno, channel=channel, load_audit=True)
            # For the purposes of testing, limit to the first 3 mjds
            ir = ir.isel(mjd=slice(3))
            # Load the look vector and transform it to ECI
            ir['look_eci'] = get_pointing_for_ir_dataset(channel, ir, attitude['Qachieved'])
            for mjd, ir in ir.groupby('mjd'):
                for pixel in range(128):
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    np.testing.assert_allclose(ir['look_eci'][pixel], pixel_look_eci, atol=1e-9,
                                               err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_pointing_for_ir_channel1_dataset(self):
        """ Test comparing the IR channel 1 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=1)

    def test_pointing_for_ir_channel2_dataset(self):
        """ Test comparing the IR channel 2 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=2)

    def test_pointing_for_ir_channel3_dataset(self):
        """ Test comparing the IR channel 3 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=3)

    def tangent_point_for_ir_dataset(self, channel: int):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR imager.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            ir = open_level1_ir(scanno=scanno, channel=channel)
            # For the purposes of testing, limit to the first 3 mjds
            ir = ir.isel(mjd=slice(3))
            # Load the look vector and transform it to ECI
            ir['look_eci'] = get_pointing_for_ir_dataset(channel, ir, attitude['Qachieved'])
            ir['position_eci'] = get_position_for_ir_dataset(ir, attitude)
            ir.merge(get_tangent_point_data_for_ir_dataset(channel, ir, attitude), inplace=True)
            for mjd, ir in ir.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                # Assert the ECI positions are the to 5 meters
                np.testing.assert_allclose(ir['position_eci'], position_eci, atol=5)
                for pixel in range(128):
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                    list(pixel_look_eci), False)[1])
                    tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])
                    if np.isfinite(float(ir['latitude'][pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(ir['latitude'][pixel]), tangentpoint.latitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(ir['longitude'][pixel]), tangentpoint.longitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(ir['altitude'][pixel]), tangentpoint.height * 1000.0, delta=5,
                                               msg=f'ScanNo: {scanno}')

    def test_tangent_point_for_ir_channel1_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 1.
        """
        self.tangent_point_for_ir_dataset(channel=1)

    def test_tangent_point_for_ir_channel2_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 2.
        """
        self.tangent_point_for_ir_dataset(channel=2)

    def test_tangent_point_for_ir_channel3_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 3.
        """
        self.tangent_point_for_ir_dataset(channel=3)

    def test_scans_in_orbit(self):
        """ Test the service returning scans in an orbit.
        """
        self.assertListEqual([scanno for scanno in get_scannos_in_orbit_below_sza(Orbit(6432), sza=90.0)],
                             [6432000, 6432001, 6432002, 6432003, 6432004, 6432005, 6432006, 6432007, 6432008, 6432009,
                              6432010, 6432011, 6432012, 6432013, 6432014, 6432015, 6432016, 6432017, 6432018, 6432019,
                              6432020, 6432021, 6432022, 6432023, 6432024, 6432025, 6432026, 6432027, 6432028, 6432063,
                              6432064, 6432065])

    def test_no_scans_in_orbit_with_attitude_file(self):
        """ Test the service returning scans in an orbit raises StopIteration when there are no scans.
        """
        with self.assertRaises(StopIteration):
            next(get_scannos_in_orbit_below_sza(Orbit(6000), sza=90.0))

    def test_no_scans_in_orbit_no_attitude_file(self):
        """ Test the service returning scans with a scan number without an attitude file
        """
        with self.assertRaises(StopIteration):
            next(get_scannos_in_orbit_below_sza(Orbit(6462), sza=90.0))

    def test_psf_fit_residuals(self):
        """ Test that output of PointSpreadFunctionFit.fit() in PythonLevel1Services matches output of the
            stand-alone project.
        """
        psf_test_dict = {6432013: {'fwhm': 0.9262741999408295,
                                   'shift': 0.05446324297262721},
                         27110030: {'fwhm': 0.9395053846503276,
                                    'shift': 0.023467222524174592},
                         4097029: {'fwhm': 0.932808490872749,
                                   'shift': 0.06667136518077217}}

        for scanno in psf_test_dict:
            psf_results = PointSpreadFunctionFit(scanno).fit()
            self.assertAlmostEqual(psf_test_dict[scanno]['fwhm'], psf_results.fwhm, places=6)
            self.assertAlmostEqual(psf_test_dict[scanno]['shift'], psf_results.shift, places=6)

    def test_spectrograph_load_psf(self):
        """ Test that the PSF corrections are applied. """
        for scanno in self.scannos:
            unapplied = open_level1_spectrograph(scanno=scanno)
            applied = open_level1_spectrograph(scanno=scanno, load_psf=True)
            # Pixels 575 and up should be the same for both wavelength and fwhm
            np.testing.assert_allclose(unapplied.wavelength[575:], applied.wavelength[575:], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)
            np.testing.assert_allclose(unapplied.fwhm[575:], applied.fwhm[575:], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)
            # The wavelengths for pixels 574 and down should differ by a constant amount that is not zero
            wave_diff = unapplied.wavelength[:575] - applied.wavelength[:575]
            self.assertNotEqual(0.0, wave_diff[0])
            np.testing.assert_allclose(wave_diff[0], wave_diff, rtol=1e-6, err_msg=f'ScanNo: {scanno}', verbose=True)
            # The fwhm for pixels 574 and down should be constant after the psf is applied.
            np.testing.assert_allclose(applied.fwhm[0], applied.fwhm[:575], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)
